﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace GameOfLife
{
    /// <summary>
    /// Like ActiveBlockBoard, but using a parellel for loop.
    /// </summary>
    class ParallelActiveBlockBoard : BytePerPixelBoard
    {
        class Block
        {
            /// <summary>
            /// Array of changes in this block:
            /// 0 = any
            /// 1 = top row
            /// 2 = left side
            /// 3 = right side
            /// 4 = bottom row
            /// 5 = top left corner
            /// 6 = top right corner
            /// 7 = bottom left corner
            /// 8 = bottom right corner
            /// </summary>
            internal readonly bool[] changes;

            /// <summary>
            /// The number of the block whose change (at each index)
            /// means that a change may occur in this block. For example,
            /// element 4 will be the number of the block above this
            /// one, as that block's "bottom row" change affects this block.
            /// </summary>
            internal readonly int[] relativeBlockIndexes;

            /// <summary>
            /// Leftmost column of block
            /// </summary>
            internal readonly int left;
            /// <summary>
            /// Rightmost column of block
            /// </summary>
            internal readonly int right;
            /// <summary>
            /// Top row of block
            /// </summary>
            internal readonly int top;
            /// <summary>
            /// Bottom row of block
            /// </summary>
            internal readonly int bottom;

            /// <summary>
            /// Index of top-left corner within byte array
            /// </summary>
            internal readonly int topLeftIndex;

            internal Block(int x, int y, int blocksPerRow, int blocksPerColumn)
            {
                // Translates a block column and row (including out of bounds)
                // into the index.
                Func<int,int,int> colRowToIndex = (col, row) => 
                    (((row+blocksPerColumn) % blocksPerColumn) * blocksPerRow) + ((col+blocksPerRow) % blocksPerRow);

                // Start with a conservative "everything's changed" 
                changes = new bool[9];
                for (int i = 0; i < changes.Length; i++)
                {
                    changes[i] = true;
                }

                relativeBlockIndexes = new int[9];
                // Changes to this block always affect the same block
                relativeBlockIndexes[0] = colRowToIndex(x, y);
                // Top row of block below us affects us
                relativeBlockIndexes[1] = colRowToIndex(x, y + 1);
                // Left side of block to right of us affects us
                relativeBlockIndexes[2] = colRowToIndex(x + 1, y);
                // Right side of block to left of us affects us
                relativeBlockIndexes[3] = colRowToIndex(x - 1, y);
                // Bottom row of block above us affects us
                relativeBlockIndexes[4] = colRowToIndex(x, y - 1);
                // Top left corner of block below and to the right of us affects us
                relativeBlockIndexes[5] = colRowToIndex(x + 1, y + 1);
                // Top right corner of block below and to the left of us affects us
                relativeBlockIndexes[6] = colRowToIndex(x - 1, y + 1);
                // Bottom left corner of block above and to the right of us affects us
                relativeBlockIndexes[7] = colRowToIndex(x + 1, y - 1);
                // Bottom right corner of block above and to the left of us affects us
                relativeBlockIndexes[8] = colRowToIndex(x - 1, y - 1);

                left = x * BlockWidth;
                top = y * BlockHeight;
                right = left + BlockWidth - 1;
                bottom = top + BlockHeight - 1;

                topLeftIndex = top * blocksPerRow*BlockWidth + left;
            }
        }

        const int BlockWidth = 25;
        const int BlockHeight = 25;

        Block[] blocks;

        int[] offsets;

        public ParallelActiveBlockBoard(int width, int height)
            : base(width, height)
        {
            CreateBlocks();
            offsets = new int[] { -width - 1, -width, -width + 1, -1, 1, width - 1, width, width + 1 };
        }

        private void CreateBlocks()
        {
            if ((Width % BlockWidth) != 0 || (Height % BlockHeight) != 0)
            {
                throw new ArgumentException("Size must be divisible by block Width/Height");
            }
            int blocksPerRow = Width / BlockWidth;
            int blocksPerColumn = Height / BlockHeight;
            blocks = new Block[blocksPerRow * blocksPerColumn];
            int blockIndex = 0;
            for (int y = 0; y < blocksPerColumn; y++)                
            {
                for (int x = 0; x < blocksPerRow; x++)
                {
                    blocks[blockIndex++] = new Block(x, y, blocksPerRow, blocksPerColumn);
                }
            }
        }

        public override void CreateNextGeneration(IBoard parent)
        {
            ParallelActiveBlockBoard parentBoard = (ParallelActiveBlockBoard)parent;

            Block[] parentBlocks = parentBoard.blocks;
            byte[] parentData = parentBoard.Data;

            Parallel.ForEach(blocks, block =>
            {
                Array.Clear(block.changes, 0, 9);
                // Work out whether or not to calculate
                bool recalculate = false;
                for (int i = 0; i < 9; i++)
                {
                    if (parentBlocks[block.relativeBlockIndexes[i]].changes[i])
                    {
                        recalculate = true;
                        break;
                    }
                }
                // Just copy if appropriate.
                if (!recalculate)
                {
                    int index = block.topLeftIndex;
                    for (int i = 0; i < BlockHeight; i++)
                    {
                        Array.Copy(parentData, index, Data, index, BlockWidth);
                        index += Width;
                    }
                }
                else
                {
                    // Recalculate this block, noting changes
                    for (int y = block.top; y <= block.bottom; y++)
                    {
                        bool topOrBottom = y == 0 || y == Height - 1;
                        int index = y * Width + block.left;
                        for (int x = block.left; x <= block.right; x++)
                        {
                            bool alive = parentBoard[x, y];

                            int neighbours = 0;
                            if (topOrBottom || x == 0 || x == Width - 1)
                            {
                                for (int dx = -1; dx < 2; dx++)
                                {
                                    for (int dy = -1; dy < 2; dy++)
                                    {
                                        // Ignore ourselves
                                        if (dx == 0 && dy == 0)
                                        {
                                            continue;
                                        }
                                        int neighbourIndex = ((y + dy + Height) % Height) * Width + (x + dx + Width) % Width;
                                        neighbours += parentData[neighbourIndex];
                                    }
                                }
                            }
                            else
                            {
                                for (int i = 0; i < offsets.Length; i++)
                                {
                                    neighbours += parentData[index + offsets[i]];
                                }
                            }
                            bool nextAlive = neighbours == 3 || (alive && neighbours == 2);
                            this[x, y] = nextAlive;
                            if (nextAlive != alive)
                            {
                                block.changes[0] = true;
                                if (y == block.top) block.changes[1] = true;
                                if (x == block.left) block.changes[2] = true;
                                if (x == block.right) block.changes[3] = true;
                                if (y == block.bottom) block.changes[4] = true;
                                if (y == block.top && x == block.left) block.changes[5] = true;
                                if (y == block.top && x == block.right) block.changes[6] = true;
                                if (y == block.bottom && x == block.left) block.changes[7] = true;
                                if (y == block.bottom && x == block.right) block.changes[8] = true;
                            }
                            index++;
                        }
                    }
                }
            });
        }
    }
}
