﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Drawing;
using System.Diagnostics;

namespace GameOfLife
{
    class Program
    {
        const int Width = 1000;
        const int Height = 500;

        IBoard[] boards;
        PictureBox imageHolder;
        volatile bool stop;

        Program(Type boardType)
        {
            boards = new IBoard[2];
            boards[0] = (IBoard) Activator.CreateInstance(boardType, Width, Height);
            boards[1] = (IBoard) Activator.CreateInstance(boardType, Width, Height);
            SetupInitialBoard(boards[0]);
            imageHolder = new PictureBox();
            imageHolder.Image = boards[0].Render();
            imageHolder.Dock = DockStyle.Fill;
        }

        void Run()
        {
            Form form = new Form();
            form.FormBorderStyle = FormBorderStyle.FixedSingle;
            form.Controls.Add(imageHolder);
            form.ClientSize = new Size(Width, Height);
            form.Text = "Game of Life";

            form.Load += StartRunning;
            form.FormClosing += delegate { stop = true; };

            Application.Run(form);
        }

        [STAThread]
        static void Main(string[] args)
        {
            Type boardType = GetBoardType();
            if (boardType != null)
            {
                new Program(boardType).Run();
            }
        }

        private static Type GetBoardType()
        {
 	        var boardTypes = (from type in typeof(Program).Assembly.GetTypes()
                              where !type.IsAbstract && typeof(IBoard).IsAssignableFrom(type)
                              orderby type.Name
                              select type).ToArray();

            for (int i = 0; i < boardTypes.Length; i++)
            {
                Console.WriteLine("{0}: {1}", (char)('A' + i), boardTypes[i].Name);
            }

            Console.WriteLine();
            Console.Write("Type to use? ");
            string input = Console.ReadLine().ToUpperInvariant();
            if (input.Length == 0)
            {
                Console.WriteLine("No type selected");
                return null;
            }
            if (input[0] < 'A' || input[0] >= 'A' + boardTypes.Length)
            {
                Console.WriteLine("Invalid type");
                return null;
            }
            return boardTypes[input[0] - 'A'];
        }

        private void StartRunning(object sender, EventArgs e)
        {
            Thread t = new Thread(BackgroundLoop);
            t.IsBackground = true;
            t.Start();
        }

        void BackgroundLoop()
        {
            Stopwatch totalStopwatch = Stopwatch.StartNew();
            Stopwatch shortStopwatch = Stopwatch.StartNew();
            int current = 0;
            int frames = 0;
            int totalFrames = 0;
            while (!stop)
            {
                int next = 1 - current;
                boards[next].CreateNextGeneration(boards[current]);
                // Comment out the next line to turn rendering off
                imageHolder.Invoke(() => imageHolder.Image = boards[next].Render());
                
                current = next;
                long elapsed = shortStopwatch.ElapsedMilliseconds;
                frames++;
                totalFrames++;
                if (elapsed >= 1000)
                {
                    Console.WriteLine("{0} fps (average={1})", (frames*1000) / elapsed,
                        (totalFrames*1000)/totalStopwatch.ElapsedMilliseconds);
                    shortStopwatch = Stopwatch.StartNew();
                    frames = 0;
                }
            }
        }

        private static void SetupInitialBoard(IBoard board)
        {
            int x = Width / 2;
            int y = Height / 2;
            board[x, y] = true;
            board[x + 1, y] = true;
            board[x, y + 1] = true;
            board[x - 1, y + 1] = true;
            board[x, y + 2] = true;
        }
    }
}
